#ifndef GRF_PARSER_H
#define GRF_PARSER_H


#include <vector>
#include <Token.H>
#include <ParserError.H>
#include <Face.H>
#include <Vertex.H>
#include <Light.H>
#include <fstream>


namespace GRF
{

class Parser
{
   public :

   Parser
   (
      const char * _fileName,
      bool         _verbose = true
   )  ;

   ~Parser() ;

   const Face * GetFace
   (
      unsigned long iFace
   )
      const ;

   const Vertex * GetVertex
   (
      unsigned long iVertex
   )
      const ;

   const Light * GetLight
   (
      unsigned long iLight
   )
      const ;

   unsigned long NVertexs
   ()
      const ;

   unsigned long NFaces
   ()
      const ;

   unsigned long NLights
   ()
      const ;

   private:

   const char *  fileName ;
   bool          verbose  ;

   struct InputStack
   {
       std::ifstream *    is ;
       struct InputStack * next ;
   }   ;


   InputStack * stack ;


   double * pCurrentFaceColorRGB ;
      // pointer to rgb-vector with current face color, NULL if none

   double currentFaceColorRGB[3] ;
      // current face color

   std::vector<Vertex> vertexs ;
   std::vector<Face>   faces   ;
   std::vector<Light>  lights  ;

   unsigned long nLights ;
      // total number of extended light sources (already processed and stored)

   unsigned long nVertexs ;
     // total number of vertex (already processed and stored)

   unsigned long nFaces   ;
     // total number of faces (already processed and stored)

   unsigned long nMeshVertexs ;
     // number of vertex in current mesh (already processed and stored)

   unsigned long meshVertexIndexBase ;
     // base or shift to add when converting from
     // mesh-relative vertex indexes to absolute vertex indexes
     // (that is: absolute vertex index of the vertex whose relative index
     //  in the current mesh is 0).

   Token  tk ; // look-ahead token

   void OpenStream( const std::string & fileName ) ;

   void CloseStream() ;

   void ReadBlock() ;
   void ReadMesh() ;
   void ReadMeshVertexs() ;
   void ReadMeshFaces() ;
   void ReadLight() ;
   void ReadVertex() ;
   void ReadFace() ;
   void ReadFace2() ;
   void ReadFacesColor() ;
   void ReadInclude() ;
   void ReadTransformBlock() ;

   void LoadNextToken() ;

   void ReadSingleToken
   (
      TokenType type
   ) ;

   void ReadSingleNumericToken() ;

   void ReadSingleTokenOpt
   (
      TokenType type
   ) ;

   void ReadVector3
   (
      double v[3]
   ) ;



} ; // class Parser
   
   
} ; // namespace GRF
   
#endif // GRF_PARSER_H
